//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MOVABLEELLIPSE_H
#define MOVABLEELLIPSE_H

#include <QGraphicsEllipseItem>
#include <QPen>
#include "CubicSplineRecon.h"
#include "EmbedderMDS.h"
#include "MainWindow.h"
#include "ResampleDialog.h"

class ResampleEllipse : public QGraphicsEllipseItem
{
public:
    ResampleEllipse(qreal x, qreal y, qreal width, qreal height, ResampleDialog &dialog, QGraphicsItem * parent = 0) :
        prevPoint(NULL), nextPoint(NULL), dialog(dialog), QGraphicsEllipseItem(x, y, width, height, parent)
    {
        setPen(Qt::NoPen);
        setFlag(QGraphicsItem::ItemIsMovable);
        setFlag(QGraphicsItem::ItemIsSelectable);
        setFlag(QGraphicsItem::ItemSendsGeometryChanges);
        setCursor(Qt::SizeVerCursor);
    }

    ResampleEllipse *getPrevPoint() const { return prevPoint; }
    ResampleEllipse *getNextPoint() const { return nextPoint; }
    void setPrevPoint(ResampleEllipse *point) { prevPoint = point; }
    void setNextPoint(ResampleEllipse *point) { nextPoint = point; }
    void notMovable() { setFlag(QGraphicsItem::ItemIsMovable, false); setCursor(Qt::ArrowCursor); }
    void setX(qreal x) { fixedX = x; QGraphicsEllipseItem::setX(x); }

protected:
    QVariant itemChange(GraphicsItemChange change, const QVariant &value)
    {
        if (change == ItemPositionChange)
        {
            QPointF newPos(fixedX, value.toPointF().y());
            if (NULL != prevPoint) newPos.setY(qMin(prevPoint->y(), newPos.y()));
            if (NULL != nextPoint) newPos.setY(qMax(nextPoint->y(), newPos.y()));
            return newPos;
        } else if (change == ItemPositionHasChanged)
        {
            dialog.updateInterpolantCurve();
        }
        return QGraphicsEllipseItem::itemChange(change, value);
    }

private:
    qreal fixedX;
    ResampleEllipse *prevPoint;
    ResampleEllipse *nextPoint;
    ResampleDialog &dialog;
};

class MovableEllipse : public QGraphicsEllipseItem
{
public:
    MovableEllipse(qreal x, qreal y, qreal width, qreal height, QGraphicsItem * parent = 0) :
        prevLine(NULL), nextLine(NULL), QGraphicsEllipseItem(x, y, width, height, parent)
    {
        setFlag(QGraphicsItem::ItemIsMovable);
        setFlag(QGraphicsItem::ItemSendsGeometryChanges);
        setFlag(QGraphicsItem::ItemIsSelectable);
    }

    QGraphicsLineItem *getNextLine() const { return nextLine; }
    QGraphicsLineItem *getPrevLine() const { return prevLine; }
    void setNextLine(QGraphicsLineItem *line)  { nextLine = line; }
    void setPrevLine(QGraphicsLineItem *line)  { prevLine = line; }

protected:
    QVariant itemChange(GraphicsItemChange change, const QVariant &value)
    {
        if (change == ItemPositionHasChanged)
        {
            QPointF point(value.toPointF());
            if (prevLine)
            {
                QLineF line(prevLine->line());
                prevLine->setLine(line.x1(), line.y1(), point.x(), point.y());
            }
            if (nextLine)
            {
                QLineF line(nextLine->line());
                nextLine->setLine(point.x(), point.y(), line.x2(), line.y2());
            }
        }
        return value;
    }

private:
    QGraphicsLineItem *prevLine;
    QGraphicsLineItem *nextLine;
};

class MovableEmbedderEllipse : public MovableEllipse {
public:
    MovableEmbedderEllipse(qreal x, qreal y, qreal width, qreal height, QGraphicsItem * parent = 0) :
        pinned(false), embedder(NULL), MovableEllipse(x, y, width, height, parent)
    {
    }

    void setEmbedder(EmbedderMDS *newEmbedder) { embedder = newEmbedder; }
    bool isPinned() const { return pinned; }
    void unpin()
    {
        pinned = false;
        setPen(QPen(Qt::black));
    }

protected:
    QVariant itemChange(GraphicsItemChange change, const QVariant &value)
    {
        if (change == ItemPositionChange && isSelected())
        {
            pinned = true;
            setPen(QPen(Qt::red));
        }
        return MovableEllipse::itemChange(change, value);
    }

    void mousePressEvent(QGraphicsSceneMouseEvent *event)
    {
        if (embedder != NULL && embedder->iteratingNow()) return;
        QGraphicsEllipseItem::mousePressEvent(event);
    }

    void mouseMoveEvent(QGraphicsSceneMouseEvent *event)
    {
        if (embedder != NULL && embedder->iteratingNow()) return;
        QGraphicsEllipseItem::mouseMoveEvent(event);
    }

    void mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
    {
        if (embedder)
        {
            if (embedder->iteratingNow()) return;
            embedder->pointUpdated(this);
        }
        QGraphicsEllipseItem::mouseReleaseEvent(event);
    }

private:
    bool pinned;
    EmbedderMDS *embedder;
};

class MovableCtlPointEllipse : public MovableEllipse {
public:
    MovableCtlPointEllipse(int pointNumber, qreal x, qreal y, qreal width, qreal height, QGraphicsItem * parent = 0) :
        pointNumber(pointNumber), spline(NULL), MovableEllipse(x, y, width, height, parent)
    {}
    void setSpline(CubicSplineRecon *newSpline) { spline = newSpline; }

protected:
    QVariant itemChange(GraphicsItemChange change, const QVariant &value)
    {
        if (change == ItemPositionHasChanged && spline != NULL)
        {
            QPointF point(value.toPointF());
            spline->setControlPoint(pointNumber, point);
            MainWindow::getInstance()->setFitSpline(false);
        }
        return MovableEllipse::itemChange(change, value);
    }

private:
    int pointNumber;
    CubicSplineRecon *spline;
};

#endif // MOVABLEELLIPSE_H
