//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef SEQUENCE_H
#define SEQUENCE_H

#include <Eigen/Core>
#include <QObject>
#include "CubicSplineRecon.h"
#include "ThinPlateRecon.h"
#include "SubspaceRecon.h"
#include "Rep.h"
class Parameters;
class EmbedderMDS;
class MovableEmbedderEllipse;
class QGraphicsItem;
class QGraphicsPathItem;
class QGraphicsScene;
class QProgressDialog;

class Sequence : public QObject
{
    Q_OBJECT

public:
    Sequence(Rep::Description description, int frames, const Parameters &parameters);
    ~Sequence();
    void addFrame(const Rep &rep, int frame);
    void denormalize(Eigen::VectorXd &vector) const;
    void denormalize(Eigen::MatrixXd &matrix) const;
    double denormalizeAndRelErr(Eigen::VectorXd &vector, int frame) const;
    void denormalizeAndRelErr(Eigen::MatrixXd &matrix, Eigen::VectorXd &errors) const;
    double relErr(const Eigen::VectorXd &vector, int frame) const;

    void evaluatePCA(const Eigen::VectorXd &coeffs, Eigen::VectorXd &dest) const;
    void evaluateThinPlate(double x, double y, Eigen::VectorXd &dest) const;
    void evaluateSpline(double t, Eigen::VectorXd &dest) const;
    void getFrame(int frame, Eigen::VectorXd &dest) const;

    void setSplineVisible(bool visible);
    void setSplineControlVisible(bool visible);
    void setNumFrames(int nFrames);
    void viewPoseSpaceConnections(bool visible);
    void viewPoseSpacePositions(bool visible);
    void setPosePointsEnabled(bool enabled);
    void fitSplineToData(bool fit);
    void uniformSplineCurve(int nControlPoints);
    void uniformResample(int nFrames);
    void addFrozenFrames(int afterFrame, int nFrames, double newFrameVal);
    void duplicateFrames(int fromFrame, int toFrame, int repetitions);

    int getSourceRows()      const { return description.rows();            }
    int getNumFrames()       const { return frames;                        }
    int getPCAComps()        const { return subspace->numComps();          }
    int getTPSSites()        const { return thinPlate->numIncludedSites(); }
    int getNumSplinePoints() const { return cubicSpline->numPoints();      }

    void switchToNewDescription(Rep::Description newDescription);

    bool embIsDirty() const { return embDirty; }
    bool tpsIsDirty() const { return tpsDirty; }
    bool usingLeastSquares() const { return usingLSRecon;       }
    bool hasFeatureMatrix()  const { return featureMatrixValid; }
    
    const QGraphicsPathItem &getGraphicsPath()  const { return *graphicsPath; }
    QGraphicsScene          *getGraphicsScene() const { return graphicsScene; }
    const SubspaceRecon     &getSubspace()      const { return *subspace;     }
    const ThinPlateRecon    &getThinPlate()     const { return *thinPlate;    }
    const EmbedderMDS       &getEmbedder()      const { return *embedder;     }
    const CubicSplineRecon  &getSpline()        const { return *cubicSpline;  }
    const Eigen::MatrixXd   &getFeatureMatrix() const { return featureMatrix; }
          CubicSplineRecon  &getSpline()              { return *cubicSpline;  }

    void initLSRecon(int numComps);
    void addLSVector(int keyframe, const Eigen::VectorXd &vector);
    void switchToLSRecon();
    
    static double validateError(double error);

public slots:
    void finalize();
    void dirtyPCA() { pcaDirty = true; }
    void dirtyEmb() { embDirty = true; }
    void dirtyTPS() { tpsDirty = true; }
    void dirtySPL() { splDirty = true; }
    void computeAll();
    void computeEmb();
    void computeTPS();
    void computeSpl();
    void unpinSelected();
    void unpinAll();
    void redrawSplineCurve();

signals:
    void newSplineCurve();

private:
    void updateSplineView();
    void updateProgress(const QString &label, int value);

    Rep::Description    description;
    Eigen::MatrixXd     featureMatrix;
    Eigen::MatrixXd     normalizedFeatures;
    std::vector<double> means;
    std::vector<double> stdDevs;
    int                 frames;

    bool                pcaDirty;
    bool                embDirty;
    bool                tpsDirty;
    bool                splDirty;

    bool                splineVisible;
    bool                splineControlVisible;
    bool                poseSpaceConnectionsVisible;
    bool                poseSpacePositionsVisible;
    bool                posePointsEnabled;

    bool                usingLSRecon;
    bool                featureMatrixValid;
    bool                fitSpline;

    const Parameters   &parameters;
    SubspaceRecon      *subspace;
    ThinPlateRecon     *thinPlate;
    CubicSplineRecon   *cubicSpline;
    EmbedderMDS        *embedder;

    QList<QGraphicsItem *> splineControlItems;
    QList<QGraphicsItem *> embedderPoints;
    QList<QGraphicsItem *> embedderLines;
    QGraphicsPathItem  *graphicsPath;
    QGraphicsScene     *graphicsScene;
    QProgressDialog    *progress;
};

#endif // SEQUENCE_H
