//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "DefGradRep.h"
#include "DefGradRepConverter.h"

DefGradRep::DefGradRep(const DefGradRepConverter &converter, const Mesh &mesh)
{
    converter.setRepFromMesh(*this, mesh);
}

Eigen::VectorXd DefGradRep::getVector() const
{
    Eigen::VectorXd vector(rots.size() + scaleShears.size() + barycenter.size());
    int startRow = 0;
    vector.segment(startRow, rots.size()       ) = Eigen::Map<const Eigen::VectorXd>(rots.data(), rots.size());
    startRow += rots.size();
    vector.segment(startRow, scaleShears.size()) = Eigen::Map<const Eigen::VectorXd>(scaleShears.data(), scaleShears.size());
    startRow += scaleShears.size();
    vector.segment(startRow, barycenter.size() ) = barycenter;
    return vector;
}

void DefGradRep::setFromVector(const Eigen::VectorXd &vec)
{
    int nFaces = (vec.size() - 3) / 9;
    int startRow = 0;
    rots = Eigen::Map<const Eigen::MatrixXd>(&vec[startRow], nFaces, 3);
    startRow += rots.size();
    scaleShears = Eigen::Map<const Eigen::MatrixXd>(&vec[startRow], nFaces, 6);
    startRow += scaleShears.size();
    barycenter = vec.segment<3>(startRow);
}

Rep::Description DefGradRep::getDescription() const
{
    Description d;
    d.parts.push_back(Description::Part("Rotations", rots.size()));
    d.parts.push_back(Description::Part("Scales/shears", scaleShears.size()));
    d.parts.push_back(Description::Part("Barycenter", 3));
    return d;
}

DefGradRep DefGradRep::operator+(const DefGradRep &other) const
{
    DefGradRep ret(*this);
    ret += other;
    return ret;
}

DefGradRep DefGradRep::operator-(const DefGradRep &other) const
{
    DefGradRep ret(*this);
    ret -= other;
    return ret;
}

DefGradRep DefGradRep::operator*(double mult) const
{
    DefGradRep ret(*this);
    ret *= mult;
    return ret;
}

DefGradRep &DefGradRep::operator+=(const DefGradRep &other)
{
    rots += other.rots;
    scaleShears += other.scaleShears;
    barycenter += other.barycenter;
    return *this;
}

DefGradRep &DefGradRep::operator-=(const DefGradRep &other)
{
    rots -= other.rots;
    scaleShears -= other.scaleShears;
    barycenter -= other.barycenter;
    return *this;
}

DefGradRep &DefGradRep::operator*=(double mult)
{
    rots *= mult;
    scaleShears *= mult;
    barycenter *= mult;
    return *this;
}
