//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "DefGradRepConverter.h"
#include <Eigen/SVD>
#include <Eigen/LU>
#include <Eigen/QR>

DefGradRepConverter::DefGradRepConverter(const Mesh &refMesh, bool standardRef)
    : standardRef(standardRef), At(Eigen::SparseMatrix<double>(refMesh.n_vertices() - 1 + refMesh.n_faces(), 3 * refMesh.n_faces()))
{
    if (!standardRef) {
        refFramesInv.reserve(refMesh.n_faces());
    }

    // Slightly larger than needed (because of the missing first column of A, i.e. row of At), but should be fine anyway
    At.reserve(9 * refMesh.n_faces());

    int f = 0;
    for (Mesh::ConstFaceIter f_it = refMesh.faces_begin();
         f_it != refMesh.faces_end(); ++f_it, ++f)
    {
        Eigen::Matrix3d refFrame, refFrameInv;
        int vInds[3];
        fillFrame(f_it, refMesh, refFrame, vInds);
        if (standardRef)
        {
            refFrameInv = Eigen::Matrix3d::Identity();
        }
        else
        {
            refFramesInv.push_back(refFrame.inverse());
            refFrameInv = refFramesInv.back();
        }

        // Build At by increasing column order for performance
        if (vInds[1] > 0)
            At.insert(vInds[1] - 1, 3 * f    ) = refFrameInv(0, 0);
        if (vInds[2] > 0)
            At.insert(vInds[2] - 1, 3 * f    ) = refFrameInv(1, 0);
        if (vInds[0] > 0)
            At.insert(vInds[0] - 1, 3 * f    ) = -refFrameInv(0, 0) - refFrameInv(1, 0) - refFrameInv(2, 0);
        At.insert(refMesh.n_vertices() - 1 + f, 3 * f    ) = refFrameInv(2, 0);

        if (vInds[1] > 0)
            At.insert(vInds[1] - 1, 3 * f + 1) = refFrameInv(0, 1);
        if (vInds[2] > 0)
            At.insert(vInds[2] - 1, 3 * f + 1) = refFrameInv(1, 1);
        if (vInds[0] > 0)
            At.insert(vInds[0] - 1, 3 * f + 1) = -refFrameInv(0, 1) - refFrameInv(1, 1) - refFrameInv(2, 1);
        At.insert(refMesh.n_vertices() - 1 + f, 3 * f + 1) = refFrameInv(2, 1);

        if (vInds[1] > 0)
            At.insert(vInds[1] - 1, 3 * f + 2) = refFrameInv(0, 2);
        if (vInds[2] > 0)
            At.insert(vInds[2] - 1, 3 * f + 2) = refFrameInv(1, 2);
        if (vInds[0] > 0)
            At.insert(vInds[0] - 1, 3 * f + 2) = -refFrameInv(0, 2) - refFrameInv(1, 2) - refFrameInv(2, 2);
        At.insert(refMesh.n_vertices() - 1 + f, 3 * f + 2) = refFrameInv(2, 2);
    }
    At.finalize();

    factorized_AtA.compute(At * At.transpose());
    assert(factorized_AtA.succeeded());
}

void DefGradRepConverter::setMeshFromRep(Mesh &targetMesh, const DefGradRep &rep) const
{
    assert(standardRef || targetMesh.n_faces() == refFramesInv.size());

    Eigen::MatrixXd B(3 * targetMesh.n_faces(), 3);
    for (unsigned f = 0; f < targetMesh.n_faces(); ++f)
    {
        Eigen::Matrix3d R = expM(rep.rots.row(f));
    
        Eigen::Matrix3d S(Eigen::Matrix3d::Zero());
        S(0, 0) = rep.scaleShears(f, 0);
        S(0, 1) = rep.scaleShears(f, 1);
        S(0, 2) = rep.scaleShears(f, 2);
        S(1, 1) = rep.scaleShears(f, 3);
        S(1, 2) = rep.scaleShears(f, 4);
        S(2, 2) = rep.scaleShears(f, 5);
    
        Eigen::Matrix3d targetDefGrad = R * S.selfadjointView<Eigen::Upper>();
        B.block<3, 3>(3 * f, 0) = targetDefGrad.transpose();
    }

    Eigen::MatrixXd x(targetMesh.n_vertices() - 1 + targetMesh.n_faces(), 3);
    #pragma omp parallel sections num_threads(3)
    {
        x.col(0) = factorized_AtA.solve(At * B.col(0));
        #pragma omp section
        x.col(1) = factorized_AtA.solve(At * B.col(1));
        #pragma omp section
        x.col(2) = factorized_AtA.solve(At * B.col(2));
    }

    Mesh::VertexIter v_it = targetMesh.vertices_begin();
    targetMesh.set_point(v_it, Mesh::Point(0, 0, 0));
    int v = 0;
    for (++v_it; v_it != targetMesh.vertices_end(); ++v_it, ++v)
    {
        Eigen::Vector3f point = x.row(v).cast<float>();
        targetMesh.set_point(v_it, fromEigenVec(point));
    }

    Mesh::Point barycenter(0, 0, 0);
    for (Mesh::ConstVertexIter v_it = targetMesh.vertices_begin();
         v_it != targetMesh.vertices_end(); ++v_it)
    {
        barycenter += targetMesh.point(v_it);
    }
    barycenter /= static_cast<float>(targetMesh.n_vertices());
    Mesh::Point correctPoint = static_cast<Mesh::Point>(fromEigenVec(rep.barycenter)) - barycenter;
    for (Mesh::VertexIter v_it = targetMesh.vertices_begin();
         v_it != targetMesh.vertices_end(); ++v_it)
    {
        targetMesh.set_point(v_it, targetMesh.point(v_it) + correctPoint);
    }
}

void DefGradRepConverter::setRepFromMesh(DefGradRep &rep, const Mesh &targetMesh) const
{
    assert(standardRef || targetMesh.n_faces() == refFramesInv.size());
    
    rep.rots        = Eigen::MatrixXd::Zero(targetMesh.n_faces(), 3);
    rep.scaleShears = Eigen::MatrixXd::Zero(targetMesh.n_faces(), 6);

    int f = 0;
    for (Mesh::ConstFaceIter f_it = targetMesh.faces_begin();
         f_it != targetMesh.faces_end(); ++f_it, ++f)
    {
        Eigen::Matrix3d tgtFrame, refFrameInv;
        fillFrame(f_it, targetMesh, tgtFrame);
        if (standardRef) {
            refFrameInv = Eigen::Matrix3d::Identity();
        } else {
            refFrameInv = refFramesInv[f];
        }

        Eigen::JacobiSVD<Eigen::Matrix3d> svd(tgtFrame * refFrameInv,
                                              Eigen::ComputeFullU | Eigen::ComputeFullV);
        Eigen::Matrix3d R(svd.matrixU() * svd.matrixV().transpose());
        rep.rots.row(f) = logM(R);

        Eigen::Matrix3d S(R.transpose() * tgtFrame * refFrameInv);
        S = (0.5 * S + 0.5 * S.transpose()).eval();

        rep.scaleShears(f, 0) = S(0, 0);
        rep.scaleShears(f, 1) = S(0, 1);
        rep.scaleShears(f, 2) = S(0, 2);
        rep.scaleShears(f, 3) = S(1, 1);
        rep.scaleShears(f, 4) = S(1, 2);
        rep.scaleShears(f, 5) = S(2, 2);
    }
    
    Mesh::Point barycenter(0, 0, 0);
    for (Mesh::ConstVertexIter v_it = targetMesh.vertices_begin();
         v_it != targetMesh.vertices_end(); ++v_it)
    {
        barycenter += targetMesh.point(v_it);
    }
    barycenter /= static_cast<float>(targetMesh.n_vertices());
    rep.barycenter = fromOMVec(barycenter).cast<double>();
}

void DefGradRepConverter::fillFrame(const Mesh::FHandle &f, const Mesh &mesh, Eigen::Matrix3d &matrix, int *vInds)
{
    Mesh::ConstFaceVertexIter fv_it = mesh.cfv_iter(f);
    if (vInds) vInds[0] = fv_it.handle().idx();
    OpenMesh::Vec3d p0(mesh.point(fv_it)); ++fv_it;
    if (vInds) vInds[1] = fv_it.handle().idx();
    OpenMesh::Vec3d p1(mesh.point(fv_it)); ++fv_it;
    if (vInds) vInds[2] = fv_it.handle().idx();
    OpenMesh::Vec3d p2(mesh.point(fv_it)); ++fv_it; assert(!fv_it);
    matrix.col(0) = fromOMVec(p1 - p0);
    matrix.col(1) = fromOMVec(p2 - p0);
    OpenMesh::Vec3d n = (p1 - p0) % (p2 - p0);
    matrix.col(2) = fromOMVec(n / sqrt(n.norm()));
}
