//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QApplication>
#include <QKeyEvent>
#include <QStyle>
#include <QStyleOptionSlider>
#include "DoubleSlider.h"

DoubleSlider::DoubleSlider(QWidget *parent)
    : QSlider(parent),
      minValue(QSlider::minimum()), // Defaults
      maxValue(QSlider::maximum()),
      doubleValue(-DBL_MAX),
      sliderPixValues(1),
      constrainInt(false)
{
    connect(this, SIGNAL(valueChanged(int)), this, SLOT(calcDoubleValue(int)));
}

void DoubleSlider::setMinimum(int min)
{
    minValue = min;
    if (minValue > doubleValue) doubleValue = minValue;
    updateQSlider();
}

int DoubleSlider::minimum() const
{
    return minValue;
}

void DoubleSlider::setMaximum(int max)
{
    maxValue = max;
    if (maxValue < doubleValue) doubleValue = maxValue;
    updateQSlider();
}

int DoubleSlider::maximum() const
{
    return maxValue;
}

void DoubleSlider::setValue(double newDoubleValue)
{
    QSlider::setValue((newDoubleValue - minValue) * sliderPixValues);
}

double DoubleSlider::value() const
{
    return doubleValue;
}

void DoubleSlider::setSnapToIntegers(bool snap)
{
    constrainInt = snap;
    resizeEvent(NULL);
}

bool DoubleSlider::snapToIntegers() const
{
    return constrainInt;
}

void DoubleSlider::setNextIntegerValue()
{
    QSlider::setValue(sliderPixValues * ((QSlider::value() + sliderPixValues) / sliderPixValues));
}

void DoubleSlider::setPrevIntegerValue()
{
    QSlider::setValue(sliderPixValues * ((QSlider::value() - 1) / sliderPixValues));
}

void DoubleSlider::resizeEvent(QResizeEvent * /* event */)
{
    if (constrainInt) updateQSlider(1);
    else
    {
        QStyleOptionSlider opt;
        initStyleOption(&opt);
        QRect gr = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderGroove, this);
        QRect sr = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderHandle, this);
        updateQSlider(gr.width() - sr.width());
    }
}

void DoubleSlider::keyPressEvent(QKeyEvent *event)
{
    switch (event->key())
    {
    case Qt::Key_PageUp:
        invertedControls() ? setPrevIntegerValue() : setNextIntegerValue();
        break;
    case Qt::Key_PageDown:
        invertedControls() ? setNextIntegerValue() : setPrevIntegerValue();
        break;
    default:
        QSlider::keyPressEvent(event);
        break;
    }
}

void DoubleSlider::updateQSlider(int newSliderPixValues)
{
    QSlider::setMinimum(0);
    if (newSliderPixValues == -1 || newSliderPixValues == sliderPixValues)
    {
        QSlider::setMaximum(sliderPixValues * (maxValue - minValue));
    }
    else
    {
        int oldSliderPixValues = sliderPixValues;
        int center = oldSliderPixValues / 2;
        sliderPixValues = newSliderPixValues;
        if (sliderPixValues > oldSliderPixValues)
        {
            QSlider::setMaximum(sliderPixValues * (maxValue - minValue));
            if (oldSliderPixValues > 0)
            {
                QSlider::setValue((QSlider::value() * sliderPixValues + center) / oldSliderPixValues);
            }
        }
        else
        {
            if (oldSliderPixValues > 0)
            {
                QSlider::setValue((QSlider::value() * sliderPixValues + center) / oldSliderPixValues);
            }
            QSlider::setMaximum(sliderPixValues * (maxValue - minValue));
        }
    }
    setTickInterval(sliderPixValues);

    if (constrainInt) setSingleStep(1);
    else setSingleStep(maxValue - minValue);
}

void DoubleSlider::calcDoubleValue(int qSliderValue)
{
    double newDoubleValue = minValue + static_cast<double>(qSliderValue) / sliderPixValues;
    
    if (doubleValue != newDoubleValue)
    {
        doubleValue = newDoubleValue;
        emit valueChanged(doubleValue);
    }
}
