//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "LinRotInvRep.h"
#include "LinRotInvRepConverter.h"

LinRotInvRep::LinRotInvRep(const LinRotInvRepConverter &converter, Mesh &mesh)
{
    converter.setRepFromMesh(*this, mesh);
}

Eigen::VectorXd LinRotInvRep::getVector() const
{
    Eigen::VectorXd vector(edgeLengths.size() + edgeAngles.size() + secondForms.size() + 6);
    int startRow = 0;
    vector.segment(startRow, edgeLengths.size()) = edgeLengths;
    startRow += edgeLengths.size();
    vector.segment(startRow, edgeAngles.size()) = edgeAngles;
    startRow += edgeAngles.size();
    vector.segment(startRow, secondForms.size()) = secondForms;
    startRow += secondForms.size();
    vector.segment<3>(startRow) = barycenter;
    startRow += 3;
    vector.segment<3>(startRow) = firstFrame;
    return vector;
}

void LinRotInvRep::setFromVector(const Eigen::VectorXd &vec)
{
    int nHalfEdges = (vec.size() - 6) / 3;
    int startRow = 0;
    edgeLengths = Eigen::Map<const Eigen::VectorXd>(&vec[startRow], nHalfEdges);
    startRow += edgeLengths.size();
    edgeAngles = Eigen::Map<const Eigen::VectorXd>(&vec[startRow], nHalfEdges);
    startRow += edgeAngles.size();
    secondForms = Eigen::Map<const Eigen::VectorXd>(&vec[startRow], nHalfEdges);
    startRow += secondForms.size();
    barycenter = vec.segment<3>(startRow);
    startRow += 3;
    firstFrame = vec.segment<3>(startRow);
}

Rep::Description LinRotInvRep::getDescription() const
{
    Description d;
    d.parts.push_back(Description::Part("Edge lengths", edgeLengths.size()));
    d.parts.push_back(Description::Part("First form angles", edgeAngles.size()));
    d.parts.push_back(Description::Part("Second discrete forms", secondForms.size()));
    d.parts.push_back(Description::Part("Barycenter", 3));
    d.parts.push_back(Description::Part("First frame", 3));
    return d;
}

LinRotInvRep LinRotInvRep::operator+(const LinRotInvRep &other) const
{
    LinRotInvRep ret(*this);
    ret += other;
    return ret;
}

LinRotInvRep LinRotInvRep::operator-(const LinRotInvRep &other) const
{
    LinRotInvRep ret(*this);
    ret -= other;
    return ret;
}

LinRotInvRep LinRotInvRep::operator*(double mult) const
{
    LinRotInvRep ret(*this);
    ret *= mult;
    return ret;
}

LinRotInvRep &LinRotInvRep::operator+=(const LinRotInvRep &other)
{
    edgeLengths += other.edgeLengths;
    edgeAngles += other.edgeAngles;
    secondForms += other.secondForms;
    barycenter += other.barycenter;
    firstFrame += other.firstFrame;
    return *this;
}

LinRotInvRep &LinRotInvRep::operator-=(const LinRotInvRep &other)
{
    edgeLengths -= other.edgeLengths;
    edgeAngles -= other.edgeAngles;
    secondForms -= other.secondForms;
    barycenter -= other.barycenter;
    firstFrame -= other.firstFrame;
    return *this;
}

LinRotInvRep &LinRotInvRep::operator*=(double mult)
{
    edgeLengths *= mult;
    edgeAngles *= mult;
    secondForms *= mult;
    barycenter *= mult;
    firstFrame *= mult;
    return *this;
}

