//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "LinRotInvRepConverter.h"
#include <Eigen/SVD>
#include <Eigen/Geometry>

LinRotInvRepConverter::LinRotInvRepConverter(const Mesh &structure) :
    interiorHalfedges(0)
{
    // To deal with boundaries properly, we should probably form an equation using the
    // triangle that is on the correct side of the relevant halfedge. For simplicity, here
    // we just form equations by traversing the boundary in only one of the two directions.
    // That means we have only 'interiorHalfedges' equations to solve when finding frames.
    for (Mesh::ConstHalfedgeIter h_it = structure.halfedges_begin();
         h_it != structure.halfedges_end(); ++h_it)
    {
        if (!structure.is_boundary(h_it)) ++interiorHalfedges;
    }

    // Set up and factorize the matrix which we'll use for finding the geometry, once we've
    // got a complete set of frames.
    Dt = Eigen::SparseMatrix<double>(structure.n_vertices(), structure.n_halfedges() + 1);
    Dt.reserve(2 * structure.n_halfedges() + 1);

    for (Mesh::ConstHalfedgeIter h_it = structure.halfedges_begin();
         h_it != structure.halfedges_end(); ++h_it)
    {
        int frame_i_idx = structure.from_vertex_handle(h_it).idx();
        int frame_j_idx = structure.to_vertex_handle(h_it).idx();
        
        Dt.insert(frame_i_idx, h_it.handle().idx()) = -1;
        Dt.insert(frame_j_idx, h_it.handle().idx()) = 1;
    }
    Dt.insert(0, structure.n_halfedges()) = 1;
    Dt.finalize();

    // Factorize
    factorized_DtD.compute(Dt * Dt.transpose());
    assert(factorized_DtD.succeeded());
}

void LinRotInvRepConverter::setMeshFromRep(Mesh &targetMesh, const LinRotInvRep &rep) const
{
    // Set up the matrix used for finding frames.
    Eigen::SparseMatrix<double> Ft(3 * targetMesh.n_vertices(), 3 * (interiorHalfedges + 1));
    Ft.reserve(12 * interiorHalfedges + 3);

    int h = 0;
    for (Mesh::HalfedgeIter h_it = targetMesh.halfedges_begin();
         h_it != targetMesh.halfedges_end(); ++h_it)
    {
        if (targetMesh.is_boundary(h_it)) continue;

        Mesh::HHandle xi = targetMesh.halfedge_handle(targetMesh.from_vertex_handle(h_it)); // x^i_1
        double angle = 0;
        while (xi != h_it.handle())
        {
            angle += rep.edgeAngles(xi.idx());
            xi = targetMesh.opposite_halfedge_handle(targetMesh.prev_halfedge_handle(xi));
        }

        OpenMesh::Vec3d xi_k(cos(angle) * rep.edgeLengths(xi.idx()),
                             sin(angle) * rep.edgeLengths(xi.idx()), 0);
        angle += rep.edgeAngles(xi.idx());
        double Li_k = rep.secondForms(xi.idx());
        OpenMesh::Vec3d xi_k_hat(xi_k[0], xi_k[1], Li_k);
        xi = targetMesh.opposite_halfedge_handle(targetMesh.prev_halfedge_handle(xi));
        OpenMesh::Vec3d xi_kp1(cos(angle) * rep.edgeLengths(xi.idx()),
                               sin(angle) * rep.edgeLengths(xi.idx()), 0);
        double Li_kp1 = rep.secondForms(xi.idx());
        OpenMesh::Vec3d xi_kp1_hat(xi_kp1[0], xi_kp1[1], Li_kp1);

        double Ln_m   = rep.secondForms(targetMesh.next_halfedge_handle(h_it).idx());
        double Ln_mp1 = rep.secondForms(targetMesh.opposite_halfedge_handle(h_it).idx());

        // Find the normal of vertex 'j' in the frame of vertex 'i'
        Eigen::Matrix<double, 2, 3> normMatrix;
        normMatrix.row(0) = fromOMVec(xi_kp1 - xi_k);
        normMatrix.row(1) = fromOMVec(xi_k);
        normMatrix(0, 2) = Li_kp1 - Li_k;
        normMatrix(1, 2) = Li_k;
        Eigen::Vector2d normMatrixVec(Ln_m, -Ln_mp1);
        Eigen::JacobiSVD< Eigen::Matrix<double, 2, 3> > svd = normMatrix.jacobiSvd(Eigen::ComputeFullU | Eigen::ComputeFullV);
        Eigen::Vector3d normal_eigen = svd.solve(normMatrixVec);
        // The SVD gives the min norm solution, so if that's greater than 1, then we have no hope of
        // finding a unit-norm solution to these equations. So we just skip this equation entirely.
        // This should never happen for discrete forms that have been derived from a mesh, but could
        // happen for arbitrary discrete forms.
        if (normal_eigen.norm() >= 1) continue;
        double alpha = sqrt(1 - normal_eigen.squaredNorm());
        Eigen::Vector3d unit_normal_eigen = normal_eigen + alpha * svd.matrixV().col(2);
        
        OpenMesh::Vec3d normal_j(fromEigenVec(unit_normal_eigen));
        OpenMesh::Vec3d xn_m_hat   = xi_kp1_hat - xi_k_hat;
        OpenMesh::Vec3d xn_m       = xn_m_hat - (xn_m_hat | normal_j) * normal_j;
        OpenMesh::Vec3d xn_mp1_hat = -xi_k_hat;
        OpenMesh::Vec3d xn_mp1     = xn_mp1_hat - (xn_mp1_hat | normal_j) * normal_j;
        
        // Check that we've picked the correct unit norm solution (i.e. the one that is consistent with
        // the relevant edge angle discrete form
        bool trueOrientation = fmod(2 * M_PI + rep.edgeAngles(targetMesh.next_halfedge_handle(h_it).idx()), 2 * M_PI) < M_PI;
        bool currOrientation = (xn_m | (xn_mp1 % normal_j)) > 0;
        if (currOrientation != trueOrientation)
        {
            unit_normal_eigen = normal_eigen - alpha * svd.matrixV().col(2);
            normal_j          = fromEigenVec(unit_normal_eigen);
            xn_m              = xn_m_hat - (xn_m_hat | normal_j) * normal_j;
            xn_mp1            = xn_mp1_hat - (xn_mp1_hat | normal_j) * normal_j;
        }
        // Haven't been able to find a reasonable normal that is consistent with the discrete forms,
        // so we'll just skip this equation. This should never happen for discrete forms that have
        // been derived from a mesh, but could happen for arbitrary discrete forms
        if (((xn_m | (xn_mp1 % normal_j)) > 0) != trueOrientation) continue;

        Mesh::HHandle xn_i = targetMesh.halfedge_handle(targetMesh.to_vertex_handle(h_it)); // x^n_1
        angle = 0;
        while (xn_i != targetMesh.next_halfedge_handle(h_it))
        {
            angle -= rep.edgeAngles(xn_i.idx());
            xn_i = targetMesh.opposite_halfedge_handle(targetMesh.prev_halfedge_handle(xn_i));
        }
        Eigen::Vector3d b_j_1 = Eigen::AngleAxisd(angle, fromOMVec(normal_j)) * fromOMVec(xn_m);
        b_j_1.normalize();
        Eigen::Vector3d b_j_2 = fromOMVec(normal_j).cross(b_j_1);
        assert(abs(b_j_2.norm() - 1) < 1e-6);

        // Now we know the left hand side of the equation related to this halfedge.
        int frame_i_idx = targetMesh.from_vertex_handle(h_it).idx();
        int frame_j_idx = targetMesh.to_vertex_handle(h_it).idx();
        
        Ft.insert(3 * frame_i_idx    , 3 * h    ) = b_j_1[0];
        Ft.insert(3 * frame_i_idx + 1, 3 * h    ) = b_j_1[1];
        Ft.insert(3 * frame_i_idx + 2, 3 * h    ) = b_j_1[2];
        Ft.insert(3 * frame_j_idx    , 3 * h    ) = -1;
        
        Ft.insert(3 * frame_i_idx    , 3 * h + 1) = b_j_2[0];
        Ft.insert(3 * frame_i_idx + 1, 3 * h + 1) = b_j_2[1];
        Ft.insert(3 * frame_i_idx + 2, 3 * h + 1) = b_j_2[2];
        Ft.insert(3 * frame_j_idx + 1, 3 * h + 1) = -1;
        
        Ft.insert(3 * frame_i_idx    , 3 * h + 2) = normal_j[0];
        Ft.insert(3 * frame_i_idx + 1, 3 * h + 2) = normal_j[1];
        Ft.insert(3 * frame_i_idx + 2, 3 * h + 2) = normal_j[2];
        Ft.insert(3 * frame_j_idx + 2, 3 * h + 2) = -1;

        // Equation counter
        ++h;
    }
    // Setting the first frame
    Ft.insert(0, 3 * interiorHalfedges    ) = 1;
    Ft.insert(1, 3 * interiorHalfedges + 1) = 1;
    Ft.insert(2, 3 * interiorHalfedges + 2) = 1;
    Ft.finalize();

    // Set up the right-hand side of the frame equations. These are basically all zero (for the difference
    // between the two frames), apart from the equation for the first frame.
    Eigen::MatrixXd framesRhs(3 * (interiorHalfedges + 1), 3);
    Eigen::Matrix3d firstFrame = expM(rep.firstFrame);
    framesRhs.setZero();
    framesRhs.row(3 * interiorHalfedges    ) = firstFrame.col(0);
    framesRhs.row(3 * interiorHalfedges + 1) = firstFrame.col(1);
    framesRhs.row(3 * interiorHalfedges + 2) = firstFrame.col(2);

    // Factorize
    Eigen::SparseLLT<Eigen::SparseMatrix<double>, Eigen::Cholmod> factorized_FtF(Ft * Ft.transpose());
    assert(factorized_FtF.succeeded());

    // Solve
    Eigen::MatrixXd frames(3 * targetMesh.n_vertices(), 3);
    #pragma omp parallel sections num_threads(3)
    {
        frames.col(0) = factorized_FtF.solve(Ft * framesRhs.col(0));
        #pragma omp section
        frames.col(1) = factorized_FtF.solve(Ft * framesRhs.col(1));
        #pragma omp section
        frames.col(2) = factorized_FtF.solve(Ft * framesRhs.col(2));
    }
    // We don't want any scaling of the frames, and it's very easy for them to end up shrinking in order
    // to minimize error in the least-squares solution. So we'll just make sure we have unit-norm frames here.
    // In fact while we're at it we may as well orthonormalize by taking the rotation part of a polar
    // decomposition. This seems to give even better results.
    #pragma omp parallel for
    for (int f = 0; f < frames.rows() / 3; ++f)
    {
        Eigen::JacobiSVD<Eigen::Matrix3d> svd(frames.block<3, 3>(3 * f, 0), Eigen::ComputeFullU | Eigen::ComputeFullV);
        frames.block<3, 3>(3 * f, 0) = svd.matrixU() * svd.matrixV().transpose();
    }


    // The right-hand side of the equations that find the geometry, given the frames
    Eigen::MatrixXd diffsRhs(targetMesh.n_halfedges() + 1, 3);
    diffsRhs.setZero();
    for (Mesh::ConstHalfedgeIter h_it = targetMesh.halfedges_begin();
        h_it != targetMesh.halfedges_end(); ++h_it)
    {
        Mesh::HHandle xi = targetMesh.halfedge_handle(targetMesh.from_vertex_handle(h_it)); // x^i_1
        double angle = 0;
        while (xi != h_it.handle())
        {
            angle += rep.edgeAngles(xi.idx());
            xi = targetMesh.opposite_halfedge_handle(targetMesh.prev_halfedge_handle(xi));
        }
        OpenMesh::Vec3d xi_k_hat(cos(angle) * rep.edgeLengths(xi.idx()),
                                 sin(angle) * rep.edgeLengths(xi.idx()),
                                 rep.secondForms(xi.idx()));
        diffsRhs.row(h_it.handle().idx()) = fromOMVec(xi_k_hat).transpose() *
                                            frames.block<3, 3>(3 * targetMesh.from_vertex_handle(h_it).idx(), 0);
    }

    // Find the final geometry
    Eigen::MatrixXd points(targetMesh.n_vertices(), 3);
    #pragma omp parallel sections num_threads(3)
    {
        points.col(0) = factorized_DtD.solve(Dt * diffsRhs.col(0));
        #pragma omp section
        points.col(1) = factorized_DtD.solve(Dt * diffsRhs.col(1));
        #pragma omp section
        points.col(2) = factorized_DtD.solve(Dt * diffsRhs.col(2));
    }

    // Give it the correct mean
    Eigen::Vector3d mean = points.colwise().sum();
    mean /= points.rows();
    points.rowwise() -= mean;
    points.rowwise() += rep.barycenter;
    
    // Set the mesh points based on the solved geometry
    for (Mesh::VertexIter v_it = targetMesh.vertices_begin();
         v_it != targetMesh.vertices_end(); ++v_it)
    {
        Eigen::Vector3f temp = points.row(v_it.handle().idx()).cast<float>();
        targetMesh.set_point(v_it, fromEigenVec(temp));
    }
}

void LinRotInvRepConverter::setRepFromMesh(LinRotInvRep &rep, Mesh &targetMesh) const
{
    // Store vertex normals with the mesh for now (could be stored separately in the
    // future in order to maintain 'targetMesh' as a const Mesh &)
    targetMesh.request_face_normals();
    targetMesh.request_vertex_normals();
    targetMesh.update_face_normals();
    
    rep.edgeLengths = Eigen::VectorXd::Zero(targetMesh.n_halfedges());
    rep.edgeAngles  = Eigen::VectorXd::Zero(targetMesh.n_halfedges());
    rep.secondForms = Eigen::VectorXd::Zero(targetMesh.n_halfedges());
    rep.barycenter  = Eigen::Vector3d(0, 0, 0);

    for (Mesh::VertexIter v_it = targetMesh.vertices_begin();
         v_it != targetMesh.vertices_end(); ++v_it)
    {
        // Calculate area-weighted normal
        Mesh::Normal n(0, 0, 0);
        for (Mesh::VertexOHalfedgeIter vh_it = targetMesh.voh_iter(v_it);
             vh_it; ++vh_it)
        {
            if (targetMesh.is_boundary(vh_it)) continue;
            
            n += targetMesh.calc_sector_area(vh_it) *
                 targetMesh.normal(targetMesh.face_handle(vh_it));
        }
        n.normalize();

        targetMesh.set_normal(v_it, n);
        rep.barycenter += fromOMVec(targetMesh.point(v_it)).cast<double>();
    }
    rep.barycenter /= targetMesh.n_vertices();

    for (Mesh::HalfedgeIter h_it = targetMesh.halfedges_begin();
         h_it != targetMesh.halfedges_end(); ++h_it)
    {
        Mesh::VHandle v0 = targetMesh.from_vertex_handle(h_it);
        Mesh::VHandle v1 = targetMesh.to_vertex_handle(h_it);
        Mesh::VHandle v2 = targetMesh.from_vertex_handle(targetMesh.prev_halfedge_handle(h_it));

        Mesh::Normal   n = targetMesh.normal(v0);
        Mesh::Point   e1 = targetMesh.point(v1) - targetMesh.point(v0);
        Mesh::Point   e2 = targetMesh.point(v2) - targetMesh.point(v0);

        rep.secondForms(h_it.handle().idx()) = (e1 | n);

        // Project onto tangent plane
        e1 -= (e1 | n) * n;
        e2 -= (e2 | n) * n;

        rep.edgeLengths(h_it.handle().idx()) = e1.norm();

        // Find angle between projected edges
        Eigen::Matrix3f frame;
        e1.normalize();
        frame.row(0) = fromOMVec(e1);
        frame.row(0).normalize();
        frame.row(2) = fromOMVec(n);
        frame.row(1) = frame.row(2).cross(frame.row(0));
        Eigen::Vector3f e2trans = frame * fromOMVec(e2);
        double angle = atan2(e2trans.y(), e2trans.x());
        // Assume that the angle is positive for now
        if (angle < 0) angle += 2 * M_PI;
        rep.edgeAngles(h_it.handle().idx()) = angle;
    }

    // Now we need to make sure we only circle each vertex once with the edge angle
    // discrete forms
    for (Mesh::VertexIter v_it = targetMesh.vertices_begin();
         v_it != targetMesh.vertices_end(); ++v_it)
    {
        double angle;
        
        while (true)
        {
            double maxAngle = -DBL_MAX;
            int maxIndex;
            angle = 0;

            for (Mesh::VertexOHalfedgeIter vh_it = targetMesh.voh_iter(v_it);
                 vh_it; ++vh_it)
            {
                angle += rep.edgeAngles(vh_it.handle().idx());
                if (angle > maxAngle)
                {
                    maxAngle = angle;
                    maxIndex = vh_it.handle().idx();
                }
            }
        
            // If there is more than one loop, assume that the largest angle should be considered
            // negative
            if (angle > 3 * M_PI) rep.edgeAngles(maxIndex) -= 2 * M_PI;
            else break;
        }
        
        assert(abs(2 * M_PI - angle) < 1e-6);
    }

    // Find and store the first frame
    Eigen::Matrix3f firstFrame;
    firstFrame.col(2) = fromOMVec(targetMesh.normal(targetMesh.vertex_handle(0)));
    firstFrame.col(0) = fromOMVec(targetMesh.point(targetMesh.to_vertex_handle(targetMesh.halfedge_handle(targetMesh.vertex_handle(0)))) - 
                                  targetMesh.point(targetMesh.vertex_handle(0)));
    firstFrame.col(0) -= firstFrame.col(0).dot(firstFrame.col(2)) * firstFrame.col(2);
    firstFrame.col(0).normalize();
    firstFrame.col(1) = firstFrame.col(2).cross(firstFrame.col(0));
    rep.firstFrame = logM(firstFrame.cast<double>());

    // Release storage for normals
    targetMesh.release_face_normals();
    targetMesh.release_vertex_normals();
}
