//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "ResampleDialog.h" 
#include "MovableEllipse.h"
#include "CubicSplineRecon.h"
#include <QGraphicsEllipseItem>

const double ResampleDialog::pointSize = 0.2;
const double ResampleDialog::scaleRatio = 100.0;

ResampleDialog::ResampleDialog(CubicSplineRecon &spline, QWidget *parent, double currentFrame)
    : QDialog(parent), spline(spline), monotone(spline.getMonotoneInterpolant())
{
    setupUi(this);
    connect(this, SIGNAL(accepted()), this, SLOT(commitToInterpolator()));

    const std::vector<double> &xs = monotone.getXs();
    const std::vector<double> &fs = monotone.getFs();

    if (currentFrame >= 0)
    {
        graphicsScene.addLine(scaleRatio * currentFrame, -100,
                              scaleRatio * currentFrame, scaleRatio * 10 * fs.back(), QPen(Qt::red));
    }

    for (unsigned i = 0; i < xs.size(); ++i)
    {
        tempMonotone.addDataValue(xs[i], fs[i]);
        ResampleEllipse *ellipse = new ResampleEllipse(scaleRatio * -pointSize,
                                                       scaleRatio * -pointSize,
                                                       2 * scaleRatio * pointSize,
                                                       2 * scaleRatio * pointSize,
                                                       *this);
        ellipse->setBrush(Qt::black);
        graphicsScene.addItem(ellipse);
        ellipse->setX(scaleRatio * xs[i]);
        ellipse->setY(scaleRatio * (fs.back() - fs[i]));
        if (i > 0)
        {
            ellipse->setPrevPoint(points.back());
            points.back()->setNextPoint(ellipse);
        }
        else
        {
            ellipse->notMovable();
        }
        points.push_back(ellipse);
    }
    points.back()->notMovable();
    tempMonotone.computeInterpolant();
    graphicsScene.addItem(&interpolantCurve);
    updateInterpolantCurve();

    graphicsView->setMouseEventsEnabled(false);
    graphicsView->setCurrentPointVisible(false);
    graphicsView->setAspectRatioMode(Qt::IgnoreAspectRatio);
    graphicsView->setScene(&graphicsScene);
}

void ResampleDialog::updateInterpolantCurve()
{
    QPainterPath newPath;

    const std::vector<double> &xs = tempMonotone.getXs();
          std::vector<double> &fs = tempMonotone.getFs();
    const std::vector<double> &ds = tempMonotone.getDerivs();

    if (xs.size() != monotone.getXs().size() ||
        xs.size() != fs.size())
    {
        return;
    }

    for (unsigned i = 1; i < fs.size() - 1; ++i)
    {
        fs[i] = fs.back() - (points[i]->y() / scaleRatio);
    }
    tempMonotone.computeInterpolant();

    newPath.moveTo(scaleRatio * xs[0], scaleRatio * (fs.back() - fs[0]));
    for (unsigned i = 0; i < xs.size() - 1; ++i)
    {
        newPath.cubicTo(scaleRatio * (2 * xs[i] + xs[i + 1]) / 3,
                        scaleRatio * (fs.back() - (fs[i] + ds[i] / 3)),
                        scaleRatio * (xs[i] + 2 * xs[i + 1]) / 3,
                        scaleRatio * (fs.back() - (fs[i + 1] - ds[i + 1] / 3)),
                        scaleRatio * xs[i + 1],
                        scaleRatio * (fs.back() - fs[i + 1]));
    }

    interpolantCurve.setPath(newPath);
}

void ResampleDialog::resizeEvent(QResizeEvent *event)
{
    const std::vector<double> &xs = monotone.getXs();
    const std::vector<double> &fs = monotone.getFs();

    double heightRatio = static_cast<double>(graphicsView->width()) / graphicsView->height();
    heightRatio /= xs.back() - xs.front();
    heightRatio *= fs.back() - fs.front();

    QRectF sceneRect(scaleRatio * (xs.front() - pointSize),
                     scaleRatio * (fs.front() - heightRatio * pointSize),
                     scaleRatio * (xs.back() + 2 * pointSize - xs.front()),
                     scaleRatio * (fs.back() + 2 * heightRatio * pointSize - fs.front()));
    graphicsScene.setSceneRect(sceneRect);

    for (std::vector<ResampleEllipse *>::iterator it = points.begin();
         it != points.end(); ++it)
    {
        (*it)->setRect(scaleRatio * -pointSize, scaleRatio * -heightRatio * pointSize,
                       scaleRatio * 2 * pointSize, scaleRatio * 2 * heightRatio * pointSize);
    }

    if (NULL != event) QDialog::resizeEvent(event);
}

void ResampleDialog::showEvent(QShowEvent * /* event */)
{
    resizeEvent(NULL);
}

void ResampleDialog::commitToInterpolator()
{
    std::vector<double> &fs = monotone.getFs();

    for (unsigned i = 1; i < fs.size() - 1; ++i)
    {
        fs[i] = fs.back() - (points[i]->y() / scaleRatio);
    }

    monotone.computeInterpolant();
}

void ResampleDialog::on_distSelButton_clicked()
{
    int selStart = -1, selEnd = -1, numSelected = 0;

    for (unsigned i = 0; i < points.size(); ++i)
    {
        if (points[i]->isSelected())
        {
            if (selStart < 0) selStart = i;
            selEnd = i;
            ++numSelected;
        }
    }

    if (selStart == -1 || selEnd == selStart) return;

    double startY = points[selStart]->y();
    double   endY = points[selEnd]->y();

    int selIx = 0;
    for (int i = selStart + 1; i < selEnd; ++i)
    {
        if (points[i]->isSelected())
        {
            ++selIx;
            points[i]->setY(startY + selIx * (endY - startY) / (numSelected - 1));
        }
    }
    selIx = 0;
    for (int i = selEnd - 1; i > selStart; --i)
    {
        if (points[i]->isSelected())
        {
            ++selIx;
            points[i]->setY(endY + selIx * (startY - endY) / (numSelected - 1));
        }
    }
}
