//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "Sequence.h"
#include "MainWindow.h"
#include "MovableEllipse.h"
#include "Parameters.h"
#include "EmbedderMDS.h"

#include <cmath>
#include <QGraphicsItem>
#include <QGraphicsScene>
#include <QProgressDialog>

Sequence::Sequence(Rep::Description description, int frames, const Parameters &parameters)
    : description(description),
      featureMatrix(Eigen::MatrixXd::Zero(description.rows(), frames)),
      normalizedFeatures(featureMatrix), frames(frames),
      pcaDirty(true), embDirty(true), tpsDirty(true), splDirty(true),
      splineVisible(true), splineControlVisible(false), poseSpaceConnectionsVisible(true), poseSpacePositionsVisible(true),
      posePointsEnabled(true), usingLSRecon(false), featureMatrixValid(true), fitSpline(true),
      parameters(parameters), subspace(NULL), thinPlate(NULL), cubicSpline(NULL), embedder(NULL),
      progress(NULL), graphicsScene(new QGraphicsScene(QRectF(0, 0, 0, 0))), graphicsPath(NULL)
{
    connect(&parameters, SIGNAL(pcaParamsChanged()), this, SLOT(dirtyPCA()));
    connect(&parameters, SIGNAL(tpsParamsChanged()), this, SLOT(dirtyTPS()));
    connect(&parameters, SIGNAL(splParamsChanged()), this, SLOT(dirtySPL()));
}

Sequence::~Sequence()
{
    if (NULL != subspace     ) delete subspace;
    if (NULL != embedder     ) delete embedder;
    if (NULL != thinPlate    ) delete thinPlate;
    if (NULL != cubicSpline  ) delete cubicSpline;
    if (NULL != graphicsScene) delete graphicsScene;
}

void Sequence::addFrame(const Rep &rep, int frame)
{
    assert(description.parts.size() > 0);
    assert(frame >= 0 && frame < frames);
    assert(rep.getDescription().rows() == featureMatrix.rows());
    featureMatrix.col(frame) = rep.getVector();
}

void Sequence::denormalize(Eigen::VectorXd &vector) const
{
    int startRow = 0, p = 0;
    for (std::vector<Rep::Description::Part>::const_iterator it = description.parts.begin();
         it != description.parts.end(); ++it, ++p)
    {
        vector.segment(startRow, it->rows) *= stdDevs[p];
        vector.segment(startRow, it->rows).array() += means[p];
        startRow += it->rows;
    }
}

void Sequence::denormalize(Eigen::MatrixXd &matrix) const
{
    int startRow = 0, p = 0;
    for (std::vector<Rep::Description::Part>::const_iterator it = description.parts.begin();
         it != description.parts.end(); ++it, ++p)
    {
        matrix.block(startRow, 0, it->rows, matrix.cols()) *= stdDevs[p];
        matrix.block(startRow, 0, it->rows, matrix.cols()).array() += means[p];
        startRow += it->rows;
    }
}

double Sequence::denormalizeAndRelErr(Eigen::VectorXd &vector, int frame) const
{
    denormalize(vector);
    return relErr(vector, frame);
}

void Sequence::denormalizeAndRelErr(Eigen::MatrixXd &matrix, Eigen::VectorXd &errors) const
{
    denormalize(matrix);
    errors = (featureMatrix - matrix).colwise().norm().array() / featureMatrix.colwise().norm().array();
    for (int i = 0; i < errors.size(); ++i) errors[i] = validateError(errors[i]);
}

double Sequence::relErr(const Eigen::VectorXd &vector, int frame) const
{
    return validateError((featureMatrix.col(frame) - vector).norm() / featureMatrix.col(frame).norm());
}

void Sequence::evaluatePCA(const Eigen::VectorXd &coeffs, Eigen::VectorXd &dest) const
{
    subspace->reconstructFrom(coeffs, dest);
    denormalize(dest);
}

void Sequence::evaluateThinPlate(double x, double y, Eigen::VectorXd &dest) const
{
    Eigen::VectorXd pcaCoeffs;
    thinPlate->evaluate(x, y, pcaCoeffs);
    evaluatePCA(pcaCoeffs, dest);
}

void Sequence::evaluateSpline(double t, Eigen::VectorXd &dest) const
{
    QPointF pt = cubicSpline->evaluate(t);
    evaluateThinPlate(pt.x(), pt.y(), dest);
}

void Sequence::getFrame(int frame, Eigen::VectorXd &dest) const
{
    dest = featureMatrix.col(frame);
}

void Sequence::setSplineVisible(bool visible)
{
    splineVisible = visible;
    if (graphicsPath != NULL)
    {
        graphicsPath->setVisible(visible);
    }
}

void Sequence::setSplineControlVisible(bool visible)
{
    splineControlVisible = visible;
    QListIterator<QGraphicsItem *> it(splineControlItems);
    while (it.hasNext())
    {
        QGraphicsItem *item = it.next();
        item->setVisible(visible);
    }
}

void Sequence::setNumFrames(int nFrames)
{
    featureMatrix.conservativeResize(Eigen::NoChange, nFrames);
    normalizedFeatures.conservativeResize(Eigen::NoChange, nFrames);
    frames = nFrames;
    dirtyPCA();
    dirtyEmb();
}

void Sequence::viewPoseSpaceConnections(bool visible)
{
    poseSpaceConnectionsVisible = visible;
    QListIterator<QGraphicsItem *> it(embedderLines);
    while (it.hasNext())
    {
        QGraphicsItem *item = it.next();
        item->setVisible(visible);
    }
}

void Sequence::viewPoseSpacePositions(bool visible)
{
    poseSpacePositionsVisible = visible;
    QListIterator<QGraphicsItem *> it(embedderPoints);
    while (it.hasNext())
    {
        QGraphicsItem *item = it.next();
        item->setVisible(visible);
    }
}

void Sequence::setPosePointsEnabled(bool enabled)
{
    posePointsEnabled = enabled;
    QListIterator<QGraphicsItem *> it(embedderPoints);
    while (it.hasNext())
    {
        QGraphicsItem *item = it.next();
        item->setEnabled(enabled);
    }
}

void Sequence::fitSplineToData(bool fit)
{
    fitSpline = fit;
    if (fitSpline)
    {
        dirtySPL();
    }
}

void Sequence::uniformSplineCurve(int nControlPoints)
{
    if (NULL != cubicSpline) delete cubicSpline;
    cubicSpline = new CubicSplineRecon(*this, graphicsScene->sceneRect().center(), nControlPoints);
    updateSplineView();
}

void Sequence::uniformResample(int nFrames)
{
    assert(NULL != cubicSpline);
    cubicSpline->uniformResample(nFrames);
    emit newSplineCurve();
}

void Sequence::addFrozenFrames(int afterFrame, int nFrames, double newFrameVal)
{
    assert(NULL != cubicSpline);
    cubicSpline->addFrozenFrames(afterFrame, nFrames, newFrameVal);
    emit newSplineCurve();
}

void Sequence::duplicateFrames(int fromFrame, int toFrame, int repetitions)
{
    assert(NULL != cubicSpline);
    cubicSpline->duplicateFrames(fromFrame, toFrame, repetitions);
    updateSplineView();
    emit newSplineCurve();
}

void Sequence::switchToNewDescription(Rep::Description newDescription)
{
    description = newDescription;
    featureMatrixValid = false;
    if (NULL != embedder) setPosePointsEnabled(false);

    assert(usingLSRecon);
    assert(NULL != subspace);
    subspace->resetSubspace();
}

void Sequence::initLSRecon(int numComps)
{
    if (NULL != subspace) delete subspace;
    subspace = new SubspaceRecon(*this);
    subspace->initSubspace(numComps);
}

void Sequence::addLSVector(int keyframe, const Eigen::VectorXd &vector)
{
    assert(NULL != subspace);

    Eigen::VectorXd normalizedVector = vector;
    int startRow = 0, p = 0;
    for (std::vector<Rep::Description::Part>::iterator it = description.parts.begin();
         it != description.parts.end(); ++it, ++p)
    {
        normalizedVector.segment(startRow, it->rows).array() -= means[p];
        normalizedVector.segment(startRow, it->rows).array() /= stdDevs[p];
        startRow += it->rows;
    }

    if (0 == keyframe)
    {
        subspace->setOffset(normalizedVector);
    }
    else
    {
        subspace->setSubspaceVector(keyframe - 1, normalizedVector);
    }
}

void Sequence::switchToLSRecon()
{
    usingLSRecon = true;
    assert(NULL != subspace);
    subspace->computeLSCoeffs(normalizedFeatures);
    dirtyTPS();
}

void Sequence::finalize()
{
    if (!featureMatrixValid) return;

    normalizedFeatures = featureMatrix;
    means.clear();
    stdDevs.clear();

    int startRow = 0;
    for (std::vector<Rep::Description::Part>::iterator it = description.parts.begin();
         it != description.parts.end(); ++it)
    {
        int n = it->rows * frames;
        double mean = normalizedFeatures.block(startRow, 0, it->rows, frames).sum() / n;
        means.push_back(mean);
        normalizedFeatures.block(startRow, 0, it->rows, frames).array() -= mean;

        double stdDev = sqrt(normalizedFeatures.block(startRow, 0, it->rows, frames).array().square().sum() / n);
        stdDevs.push_back(stdDev);
        normalizedFeatures.block(startRow, 0, it->rows, frames) /= stdDev;

        startRow += it->rows;
    }
}

void Sequence::computeAll()
{
    if (pcaDirty)
    {
        tpsDirty = true;

        if (!usingLSRecon)
        {
            if (NULL != subspace) delete subspace;
            updateProgress("Computing shape signal...", 0);
            subspace = new SubspaceRecon(*this);
            subspace->computePCA(normalizedFeatures, parameters.PCATol(), progress);
        }

        pcaDirty = false;
    }

    computeEmb();
}

void Sequence::computeEmb()
{
    if (embDirty)
    {
        tpsDirty = true;
        if (NULL != embedder) delete embedder;

        embedder = new EmbedderMDS(normalizedFeatures, *this);
        embedder->iterateToConvergence();
        connect(embedder, SIGNAL(embeddingChanged()), this, SLOT(dirtyTPS()));

        // Clean scene
        QListIterator<QGraphicsItem *> it(embedderLines);
        while (it.hasNext())
        {
            QGraphicsItem *item = it.next();
            graphicsScene->removeItem(item);
            delete item;
        }
        embedderLines.clear();
        it = QListIterator<QGraphicsItem *>(embedderPoints);
        while (it.hasNext())
        {
            QGraphicsItem *item = it.next();
            graphicsScene->removeItem(item);
            delete item;
        }
        embedderPoints.clear();

        MovableEmbedderEllipse *prevEllipse = NULL;
        double sceneSize = qMax(embedder->boundingRect().height(), embedder->boundingRect().width());
        for (int frame = 0; frame < frames; ++frame)
        {
            MovableEmbedderEllipse *ellipse = new MovableEmbedderEllipse(-sceneSize / 150, -sceneSize / 150,
                                                                         sceneSize / 75, sceneSize / 75);
            ellipse->setPen(QPen(Qt::black));
            ellipse->setBrush(QBrush(Qt::white));
            embedder->setPoint(frame, ellipse);
            ellipse->setZValue(1);
            graphicsScene->addItem(ellipse);
            embedderPoints.append(ellipse);

            if (frame > 0)
            {
                QGraphicsLineItem *line = graphicsScene->addLine(embedder->frameX(frame - 1),
                                                                 embedder->frameY(frame - 1),
                                                                 embedder->frameX(frame),
                                                                 embedder->frameY(frame),
                                                                 QPen(Qt::black));
                embedderLines.append(line);
                ellipse->setPrevLine(line);
                prevEllipse->setNextLine(line);
            }

            prevEllipse = ellipse;
        }
        viewPoseSpacePositions(poseSpacePositionsVisible);
        viewPoseSpaceConnections(poseSpaceConnectionsVisible);
        setPosePointsEnabled(posePointsEnabled);

        embDirty = false;
    }

    computeTPS();
}

void Sequence::computeTPS()
{
    if (tpsDirty)
    {
        splDirty = true;

        if (NULL != thinPlate) delete thinPlate;

        updateProgress("Computing pose signal...", getNumFrames());
        thinPlate = new ThinPlateRecon(parameters.TPSTol(), *this, featureMatrixValid, parameters.TPSSmoothness(), progress);

        tpsDirty = false;
    }

    computeSpl();
}

void Sequence::computeSpl()
{
    if (splDirty && (fitSpline || NULL == cubicSpline))
    {
        if (NULL != cubicSpline) delete cubicSpline;

        Eigen::MatrixXd targetPoints = Eigen::MatrixXd::Zero(getNumFrames(), 2);
        for (int f = 0; f < getNumFrames(); ++f)
        {
            targetPoints(f, 0) = embedder->frameX(f);
            targetPoints(f, 1) = embedder->frameY(f);
        }

        updateProgress("Computing time signal...", 2 * getNumFrames() - 3);
        cubicSpline = new CubicSplineRecon(targetPoints, parameters.SPLTol(), *this, parameters.SPLAlpha(), progress);

        updateSplineView();
        progress->setValue(3 * getNumFrames() - 5);

        splDirty = false;
    }

    delete progress;
    progress = NULL;
}

void Sequence::unpinSelected()
{
    if (NULL == embedder) return;
    embedder->unpinSelected();
}

void Sequence::unpinAll()
{
    if (NULL == embedder) return;
    embedder->unpinAll();
}

void Sequence::updateSplineView()
{
    updateProgress("Drawing time signal...", 3 * getNumFrames() - 6);
    redrawSplineCurve();

    QListIterator<QGraphicsItem *> it(splineControlItems);
    while (it.hasNext())
    {
        QGraphicsItem *item = it.next();
        graphicsScene->removeItem(item);
    }
    graphicsScene->setSceneRect(QRectF());
    splineControlItems.clear();
    const Eigen::MatrixXd &controlPoints = cubicSpline->getControlPoints();
    MovableCtlPointEllipse *prevEllipse = NULL;
    double sceneSize = qMax(embedder->boundingRect().height(), embedder->boundingRect().width());
    for (int point = 0; point < controlPoints.rows(); ++point)
    {
        MovableCtlPointEllipse *ellipse = new MovableCtlPointEllipse(point, -sceneSize / 150, -sceneSize / 150,
                                                                     sceneSize / 75, sceneSize / 75);
        ellipse->setPen(QPen(Qt::blue));
        ellipse->setBrush(QBrush(Qt::white));
        ellipse->setZValue(1);
        ellipse->setX(controlPoints(point, 0));
        ellipse->setY(controlPoints(point, 1));
        ellipse->setSpline(cubicSpline);
        graphicsScene->addItem(ellipse);
        splineControlItems.append(ellipse);

        if (point > 0)
        {
            QGraphicsLineItem *line = graphicsScene->addLine(controlPoints(point - 1, 0),
                                                             controlPoints(point - 1, 1),
                                                             controlPoints(point, 0),
                                                             controlPoints(point, 1),
                                                             QPen(Qt::blue));
            splineControlItems.append(line);
            ellipse->setPrevLine(line);
            prevEllipse->setNextLine(line);
        }

        prevEllipse = ellipse;
    }
    setSplineControlVisible(splineControlVisible);
    
    // Fix the scene rectangle so that the evaluation point can't change it
    graphicsScene->setSceneRect(graphicsScene->sceneRect());

    emit newSplineCurve();
}

void Sequence::updateProgress(const QString &label, int value)
{
    if (NULL == progress)
    {
        progress = new QProgressDialog(label, QString(), value, 3 * getNumFrames() - 5, MainWindow::getInstance());
        progress->setWindowModality(Qt::WindowModal);
        progress->setWindowTitle("Find signals");
        progress->show(); progress->repaint();
    }
    else
    {
        progress->setLabelText(label);
        progress->setValue(value);
    }
}

void Sequence::redrawSplineCurve()
{
    if (graphicsPath != NULL)
    {
        graphicsScene->removeItem(graphicsPath);
        graphicsPath = NULL;
    }
    QPainterPath path;
    const std::vector<double> &knots = cubicSpline->getKnots();
    QPointF c0 = cubicSpline->directEvaluate(knots[2]);
    path.moveTo(c0);
    QPointF d0 = cubicSpline->directEvaluate(knots[2], 1);
    for(unsigned i = 3; i <= knots.size() - 3; ++i)
    {
        const QPointF c3 = cubicSpline->directEvaluate(knots[i]);
        const QPointF d3 = cubicSpline->directEvaluate(knots[i], 1);
        double spanSize = knots[i] - knots[i - 1];
        path.cubicTo(c0 + spanSize * d0 / 3, c3 - spanSize * d3 / 3, c3);
        c0 = c3;
        d0 = d3;
    }
    graphicsPath = graphicsScene->addPath(path);
    setSplineVisible(splineVisible);
}

double Sequence::validateError(double error)
{
    if (_isnan(error)) return std::numeric_limits<double>::infinity();
    else return error;
}
